/*
 *	aegis - project change supervisor
 *	Copyright (C) 2003 Peter Miller;
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 * MANIFEST: functions to manipulate files
 */

#include <ac/stdio.h>
#include <ac/string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <change.h>
#include <dir_stack.h>
#include <file.h>
#include <get/project/file.h>
#include <http.h>
#include <now.h>
#include <os.h>
#include <project.h>
#include <project/file.h>
#include <str_list.h>


static void
emit_file(string_ty *filename, struct stat *st)
{
    string_ty       *to;

    /*
     * Print the output header.
     */
    printf("Content-Length: %ld\n", st->st_size);
    http_content_type_header(filename);

    /*
     * Now copy the file contents to stdout.
     */
    to = str_from_c("");
    copy_whole_file(filename, to, 0);
    str_free(to);
}


static string_ty *
hunt(string_list_ty *gizzards, const char *name)
{
    size_t          j;
    string_ty       *s;

    for (j = 0; j < gizzards->nstrings; ++j)
    {
	s = gizzards->string[j];
	if (0 == strcasecmp(s->str_text, name))
	    return s;
    }
    return 0;
}


static string_ty *
cat4(const char *s1, string_ty *s2, string_ty *s3, string_ty *s4)
{
    string_ty       *temp1;
    string_ty       *temp2;
    string_ty       *temp3;
    string_ty       *result;

    temp1 = str_from_c(s1);
    temp2 = os_path_cat(temp1, s2);
    temp3 = os_path_cat(temp2, s3);
    result = os_path_cat(temp3, s4);
    str_free(temp1);
    str_free(temp2);
    str_free(temp3);
    return result;
}


static const char *
mode_string(int mode)
{
    static char     buffer[11];

    buffer[0] = '?';
    switch (mode & S_IFMT)
    {
    case S_IFDIR: buffer[0] = 'd'; break;
    case S_IFREG: buffer[0] = '-'; break;
    case S_IFLNK: buffer[0] = 'l'; break;
    }
    buffer[1] = (mode & 0400) ? 'r' : '-';
    buffer[2] = (mode & 0200) ? 'w' : '-';
    buffer[3] = "-xSs"[((mode >> 6) & 1) | ((mode >> 10) & 2)];
    buffer[4] = (mode & 0040) ? 'r' : '-';
    buffer[5] = (mode & 0020) ? 'w' : '-';
    buffer[6] = "-xSs"[((mode >> 3) & 1) | ((mode >> 9) & 2)];
    buffer[7] = (mode & 0004) ? 'r' : '-';
    buffer[8] = (mode & 0002) ? 'w' : '-';
    buffer[9] = "-xTt"[(mode & 1) | ((mode >> 8) & 2)];
    buffer[10] = 0;
    return buffer;
}


static void
emit_dir(change_ty *cp, string_list_ty *search_path, string_ty *filename)
{
    string_list_ty  gizzards;
    size_t          j;
    time_t          time_split;
    string_ty       *idx;

    /*
     * Read the directory contents.
     */
    dir_stack_readdir(search_path, filename, &gizzards);
    string_list_sort(&gizzards);

    idx = hunt(&gizzards, "index.html");
    if (!idx)
	idx = hunt(&gizzards, "index.htm");
    if (!idx)
	idx = hunt(&gizzards, "welcome.html");
    if (!idx)
	idx = hunt(&gizzards, "welcome.htm");
    if (idx)
    {
	string_ty       *path;
	string_ty       *abs_path;
	struct stat     st;

	path = os_path_cat(filename, idx);
	abs_path = dir_stack_find(search_path, 0, path, &st, 0, 1);
	str_free(path);
	if (abs_path)
	{
	    if (cp->bogus)
	    {
		idx =
		    cat4
		    (
			http_script_name(),
			project_name_get(cp->pp),
			filename,
			idx
		    );
	    }
	    else
	    {
		string_ty       *s;

		s =
		    str_format
		    (
			"%S.C%3.3ld",
			project_name_get(cp->pp),
			cp->number
		    );
		idx = cat4(http_script_name(), s, filename, idx);
		str_free(s);
	    }
	    printf("Location: ");
	    html_escape_string(idx);
	    str_free(idx);
	    printf("\n\n");

	    string_list_destructor(&gizzards);
	    return;
	}
    }

    /*
     * Print the output header.
     */
    html_header(cp->pp);
    printf("<title>Project ");
    html_encode_string(project_name_get(cp->pp));
    if (!cp->bogus)
	printf(",\nChange %ld", cp->number);
    printf(",\nDirectory ");
    html_encode_string(filename);
    printf("</title></head>\n<body><h1 align=center>\n");
    emit_change(cp);
    printf(",<br>\nDirectory \"");
    /* FIXME: link each part, too */
    html_encode_string(filename);
    printf("\"\n</h1>\n");

    printf("<table>\n");
    printf("<tr><th>Mode</th><th colspan=2>Date</th><th>Size</th>");
    printf("<th>Filename</th></tr>\n");

    time_split = now() - 6*30*24*60*60;
    for (j = 0; j < gizzards.nstrings; ++j)
    {
	string_ty       *deeper;
	string_ty       *ref;
	string_ty       *abs;
	struct stat     st;

	deeper = os_path_cat(filename, gizzards.string[j]);
	abs = dir_stack_find(search_path, 0, deeper, &st, 0, 1);
	if (!abs)
	    continue;
	str_free(abs);

	printf("<tr>\n");

	printf("<td valign=top><tt>");
	html_encode_charstar(mode_string(st.st_mode));
	printf("</tt></td>\n");

	printf("<td valign=top>");
	{
	    struct tm *tm;
	    char buffer[100];

	    tm = localtime(&st.st_mtime);
	    strftime(buffer, sizeof(buffer), "%b %e", tm);
	    html_encode_charstar(buffer);
	    printf("</td><td valign=top>");
	    if (st.st_mtime < time_split)
		strftime(buffer, sizeof(buffer), "%Y", tm);
	    else
		strftime(buffer, sizeof(buffer), "%H:%M", tm);
	    html_encode_charstar(buffer);
	}
	printf("</td>\n");

	printf("<td valign=top align=right>");
	printf("%ld", st.st_size);
	printf("</td>\n");

	printf("<td valign=top>\n<a href=\"");
	if (cp->bogus)
	{
	    ref =
		cat4
		(
		    http_script_name(),
		    project_name_get(cp->pp),
		    filename,
		    gizzards.string[j]
		);
	}
	else
	{
	    string_ty       *s;

	    s = str_format("%S.C%03ld", project_name_get(cp->pp), cp->number);
	    ref = cat4(http_script_name(), s, filename, gizzards.string[j]);
	    str_free(s);
	}
	html_escape_string(ref);
	str_free(ref);
	if ((st.st_mode & S_IFMT) == S_IFDIR)
	    printf("/");
	printf("\">\n");
	html_encode_string(gizzards.string[j]);
	printf("</a>");
	if ((st.st_mode & S_IFMT) == S_IFDIR)
	    printf("/");
	printf("\n</td>");
	printf("</tr>\n");
    }

    printf("</table>\n");
    printf("Listed %lu files.\n", (unsigned long)gizzards.nstrings);
    string_list_destructor(&gizzards);
    html_footer();
}


void
get_file(change_ty *cp, string_list_ty *search_path, string_ty *filename)
{
    string_ty       *absolute_path;
    struct stat     st;

    os_become_orig();
    absolute_path = dir_stack_find(search_path, 0, filename, &st, 0, 1);
    if (!absolute_path)
	http_fatal("No such file.");
    switch (st.st_mode & S_IFMT)
    {
    case S_IFDIR:
	emit_dir(cp, search_path, filename);
	break;

    case S_IFREG:
	emit_file(absolute_path, &st);
	break;

    default:
	http_fatal("Not an appropriate file.");
    }
    str_free(absolute_path);
    os_become_undo();
}


void
get_project_file(project_ty *pp, string_ty *filename, string_list_ty *modifier)
{
    string_list_ty  search_path;
    change_ty       *cp;

    string_list_constructor(&search_path);
    project_search_path_get(pp, &search_path, 0);
    cp = change_bogus(pp);
    get_file(cp, &search_path, filename);
    change_free(cp);
    string_list_destructor(&search_path);
}
